@@ -2,29 +2,32 @@
 Module for the SpecObjs and SpecObj classes
 
 .. include common links, assuming primary doc root is up one directory
-.. include:: ../links.rst
+.. include:: ../include/links.rst
 """
 import os
 import re
+from typing import List
 
 import numpy as np
 
 from astropy import units
 from astropy.io import fits
 from astropy.table import Table
+from astropy.time import Time
 
 from pypeit import msgs
 from pypeit import specobj
-from pypeit.io import initialize_header
+from pypeit import io
 from pypeit.spectrographs.util import load_spectrograph
 from pypeit.core import parse
 from pypeit.images import detector_container
 from pypeit import slittrace
+from pypeit import utils
 
 from IPython import embed
 
 
-class SpecObjs(object):
+class SpecObjs:
     """
     Object to hold a set of :class:`~pypeit.specobj.SpecObj` objects
 
@@ -45,10 +48,13 @@ class SpecObjs(object):
     """
     version = '1.0.0'
 
+    #TODO JFH This method only populates some of the underlying specobj attributes, for example RA and DEC are not
+    # getting set. We should do our best to populate everything that got written out to the file. This is part of having
+    # a rigid data model.
     @classmethod
-    def from_fitsfile(cls, fits_file, det=None):
+    def from_fitsfile(cls, fits_file, det=None, chk_version=True):
         """
-        Instantiate from a FITS file
+        Instantiate from a spec1d FITS file
 
         Also tag on the Header
 
@@ -56,17 +62,20 @@ class SpecObjs(object):
             fits_file (str):
             det (int, optional):
                 Only load SpecObj matching this det value
+            chk_version (:obj:`bool`):
+                If False, allow a mismatch in datamodel to proceed
 
         Returns:
             specobsj.SpecObjs
 
         """
         # HDUList
-        hdul = fits.open(fits_file)
+        hdul = io.fits_open(fits_file)
         # Init
         slf = cls()
         # Add on the header
         slf.header = hdul[0].header
+        # Keep track of HDUList for closing later
 
         detector_hdus = {}
         # Loop for Detectors first as we need to add these to the objects
@@ -77,7 +86,7 @@ class SpecObjs(object):
         for hdu in hdul[1:]:
             if 'DETECTOR' in hdu.name:
                 continue
-            sobj = specobj.SpecObj.from_hdu(hdu)
+            sobj = specobj.SpecObj.from_hdu(hdu, chk_version=chk_version)
             # Restrict on det?
             if det is not None and sobj.DET != det:
                 continue
@@ -87,8 +96,10 @@ class SpecObjs(object):
             # Append
             slf.add_sobj(sobj)
         # Return
+        hdul.close()
         return slf
 
+
     def __init__(self, specobjs=None, header=None):
 
         # Only two attributes are allowed for this Object -- specobjs, header
@@ -100,6 +111,7 @@ class SpecObjs(object):
             self.specobjs = specobjs
 
         self.header = header if header is not None else None
+        self.hdul = None
 
         # Turn off attributes from here
         #   Anything else set will be on the individual specobj objects in the specobjs array
@@ -207,6 +219,7 @@ class SpecObjs(object):
         meta_spec['PYP_SPEC'] = self.header['PYP_SPEC']
         meta_spec['PYPELINE'] = self[0].PYPELINE
         meta_spec['DET'] = detector
+        meta_spec['DISPNAME'] = self.header['DISPNAME']
         # Return
         if self[0].PYPELINE in ['MultiSlit', 'IFU'] and self.nobj == 1:
             meta_spec['ECH_ORDERS'] = None
@@ -267,9 +280,9 @@ class SpecObjs(object):
                     ind = (self.ECH_FRACPOS == uni_objid[iobj]) & (self.ECH_ORDER == uni_order[iord])
                     spec = self[ind]
                     # Grab SNR
-                    if self.OPT_COUNTS[0] is not None:
+                    if spec[0].OPT_COUNTS is not None:
                         SNR[iord, iobj] = np.median(spec[0].OPT_COUNTS*np.sqrt(spec[0].OPT_COUNTS_IVAR))
-                    elif self.BOX_COUNTS[0] is not None:
+                    elif spec[0].BOX_COUNTS is not None:
                         SNR[iord, iobj] = np.median(spec[0].BOX_COUNTS * np.sqrt(spec[0].BOX_COUNTS_IVAR))
                     else:
                         return None
@@ -331,6 +344,30 @@ class SpecObjs(object):
             self.remove_sobj(index)
 
 
+    def make_neg_pos(self):
+        """
+        Purge negative objects from specobjs for IR reductions
+
+        """
+        # Assign the sign and the objids
+        if self.nobj > 0:
+            if self[0].PYPELINE == 'Echelle':
+                index = self.ECH_OBJID < 0
+            elif self[0].PYPELINE == 'MultiSlit':
+                index = self.OBJID < 0
+            elif self[0].PYPELINE == 'IFU':
+                index = self.OBJID < 0
+            else:
+                msgs.error("Should not get here")
+            try:
+                self[index].OPT_COUNTS *= -1
+            except (TypeError,ValueError):
+                pass
+            try:
+                self[index].BOX_COUNTS *= -1
+            except (TypeError,ValueError):
+                pass
+
     def slitorder_indices(self, slitorder):
         """
         Return the set of indices matching the input slit/order
@@ -372,7 +409,6 @@ class SpecObjs(object):
             msgs.error("The '{0:s}' PYPELINE is not defined".format(self[0].PYPELINE))
         return indx
 
-
     def slitorder_objid_indices(self, slitorder, objid):
         """
         Return the set of indices matching the input slit/order and the input objid
@@ -423,6 +459,30 @@ class SpecObjs(object):
         # Do it
         self.specobjs = self.specobjs[msk]
 
+    def ready_for_fluxing(self):
+        # Fluxing
+        required_header = ['EXPTIME', 'AIRMASS']  # These are sufficient to apply a sensitivity function
+        required_header += ['DISPNAME', 'PYP_SPEC', 'RA', 'DEC']  # These are to generate one
+        required_for_fluxing = ['_WAVE', '_COUNTS', '_IVAR']
+
+        chk = True
+        # Check header
+        for key in required_header:
+            chk &= key in self.header
+
+        for sobj in self.specobjs:
+            sub_box, sub_opt = True, True
+            if sobj is not None:
+                # Only need one of these but need them all
+                for item in required_for_fluxing:
+                    if hasattr(sobj, 'BOX'+item):
+                        sub_box &= True
+                    if hasattr(sobj, 'OPT'+item):
+                        sub_opt &= True
+                # chk
+                chk &= (sub_box or sub_opt)
+        return chk
+
     def copy(self):
         """
         Generate a copy of self
@@ -495,13 +555,13 @@ class SpecObjs(object):
         return len(self.specobjs)
 
     def write_to_fits(self, subheader, outfile, overwrite=True, update_det=None,
-                      slitspatnum=None, debug=False):
+                      slitspatnum=None, history=None, debug=False):
         """
         Write the set of SpecObj objects to one multi-extension FITS file
 
         Args:
-            outfile (str):
             subheader (:obj:`dict`):
+            outfile (str):
             overwrite (bool, optional):
             slitspatnum (:obj:`str` or :obj:`list`, optional):
                 Restricted set of slits for reduction
@@ -537,9 +597,14 @@ class SpecObjs(object):
             _specobjs = self.specobjs
 
         # Build up the Header
-        header = initialize_header(primary=True)
+        header = io.initialize_header(primary=True)
         for key in subheader.keys():
-            header[key.upper()] = subheader[key]
+            if key.upper() == 'HISTORY':
+                if history is None:
+                    for line in str(subheader[key.upper()]).split('\n'):
+                        header[key.upper()] = line
+            else:
+                header[key.upper()] = subheader[key]
 
         # Init
         prihdu = fits.PrimaryHDU()
@@ -550,6 +615,10 @@ class SpecObjs(object):
         prihdu.header['DMODCLS'] = (self.__class__.__name__, 'Datamodel class')
         prihdu.header['DMODVER'] = (self.version, 'Datamodel version')
 
+        # Add history
+        if history is not None:
+            history.write_to_header(prihdu.header)
+
         detector_hdus = {}
         nspec, ext = 0, 0
         # Loop on the SpecObj objects
@@ -596,12 +665,12 @@ class SpecObjs(object):
         prihdu.header['NSPEC'] = nspec
 
         # Code versions
-        initialize_header(hdr=prihdu.header)
+        io.initialize_header(hdr=prihdu.header)
 
         # Finish
         hdulist = fits.HDUList(hdus)
         if debug:
-            import pdb; pdb.set_trace()
+            embed()
         hdulist.writeto(outfile, overwrite=overwrite)
         msgs.info("Wrote 1D spectra to {:s}".format(outfile))
         return
@@ -615,16 +684,19 @@ class SpecObjs(object):
             pypeline (:obj:`str`): PypeIt pipeline mode
         """
         # TODO -- Deal with update_det
-        slits, names, spat_pixpos, spat_fracpos, boxsize, opt_fwhm, s2n = [], [], [], [], [], [], []  # Lists for a Table
+        # Lists for a Table
+        slits, names, maskdef_id, objname, objra, objdec, spat_pixpos, spat_fracpos, boxsize, opt_fwhm, s2n = \
+            [], [], [], [], [], [], [], [], [], [], []
+        wave_rms = []
+        maskdef_extract = []
+        manual_extract = []
         # binspectral, binspatial = parse.parse_binning(binning)
         for specobj in self.specobjs:
             det = specobj.DET
             if specobj is None:
                 continue
             # Detector items
-            #binspectral, binspatial = parse.parse_binning(sci_dict[det]['detector'].binning)
             binspectral, binspatial = parse.parse_binning(specobj.DETECTOR.binning)
-            #platescale = sci_dict[det]['detector'].platescale
             platescale = specobj.DETECTOR.platescale
             # Append
             spat_pixpos.append(specobj.SPAT_PIXPOS)
@@ -632,14 +704,17 @@ class SpecObjs(object):
                 spat_fracpos.append(specobj.SPAT_FRACPOS)
                 slits.append(specobj.SLITID)
                 names.append(specobj.NAME)
+                wave_rms.append(specobj.WAVE_RMS)
             elif pypeline == 'IFU':
                 spat_fracpos.append(specobj.SPAT_FRACPOS)
                 slits.append(specobj.SLITID)
                 names.append(specobj.NAME)
+                wave_rms.append(specobj.WAVE_RMS)
             elif pypeline == 'Echelle':
                 spat_fracpos.append(specobj.ECH_FRACPOS)
                 slits.append(specobj.ECH_ORDER)
                 names.append(specobj.ECH_NAME)
+                wave_rms.append(specobj.WAVE_RMS)
             # Boxcar width
             if specobj.BOX_RADIUS is not None:
                 slit_pix = 2.0 * specobj.BOX_RADIUS
@@ -656,8 +731,7 @@ class SpecObjs(object):
 
             # Optimal profile (FWHM)
             # S2N -- default to boxcar
-            if specobj.FWHMFIT is not None:
-                # opt_fwhm.append(np.median(specobj.FWHMFIT)* binspatial*spectrograph.detector[specobj.DET-1]['platescale'])
+            if specobj.FWHMFIT is not None and specobj.OPT_COUNTS is not None:
                 opt_fwhm.append(np.median(specobj.FWHMFIT) * binspatial * platescale)
                 # S2N -- optimal
                 ivar = specobj.OPT_COUNTS_IVAR
@@ -671,6 +745,18 @@ class SpecObjs(object):
                     ivar = specobj.BOX_COUNTS_IVAR
                     is2n = np.median(specobj.BOX_COUNTS * np.sqrt(ivar))
                 s2n.append(is2n)
+            # Manual extraction?
+            manual_extract.append(specobj.hand_extract_flag)
+            # Slitmask info
+            if specobj.MASKDEF_ID is not None:
+                maskdef_id.append(specobj.MASKDEF_ID)
+            if specobj.MASKDEF_OBJNAME is not None:
+                objname.append(specobj.MASKDEF_OBJNAME)
+            if specobj.RA is not None:
+                objra.append(specobj.RA)
+                objdec.append(specobj.DEC)
+            if specobj.MASKDEF_EXTRACT is not None:
+                maskdef_extract.append(specobj.MASKDEF_EXTRACT)
 
         # Generate the table, if we have at least one source
         if len(names) > 0:
@@ -685,6 +771,16 @@ class SpecObjs(object):
                 obj_tbl['order'] = slits
                 obj_tbl['order'].format = 'd'
             obj_tbl['name'] = names
+            if len(maskdef_id) > 0:
+                obj_tbl['maskdef_id'] = maskdef_id
+                obj_tbl['maskdef_id'].format = 'd'
+            if len(objname) > 0:
+                obj_tbl['objname'] = objname
+            if len(objra) > 0:
+                obj_tbl['objra'] = objra
+                obj_tbl['objra'].format = '.5f'
+                obj_tbl['objdec'] = objdec
+                obj_tbl['objdec'].format = '.5f'
             obj_tbl['spat_pixpos'] = spat_pixpos
             obj_tbl['spat_pixpos'].format = '.1f'
             obj_tbl['spat_fracpos'] = spat_fracpos
@@ -697,9 +793,62 @@ class SpecObjs(object):
             obj_tbl['opt_fwhm'].unit = units.arcsec
             obj_tbl['s2n'] = s2n
             obj_tbl['s2n'].format = '.2f'
+            # is this a forced extraction at the expected position from slitmask design?
+            if len(maskdef_extract) > 0:
+                obj_tbl['maskdef_extract'] = maskdef_extract
+            # only if manual extractions exist, print this
+            if np.any(manual_extract):
+                obj_tbl['manual_extract'] = manual_extract
+
+            # Wavelengths
+            obj_tbl['wv_rms'] = wave_rms
+            obj_tbl['wv_rms'].format = '.3f'
             # Write
             obj_tbl.write(outfile,format='ascii.fixed_width', overwrite=True)
 
+    def get_extraction_groups(self, model_full_slit=False) -> List[List[int]]:
+        """
+        Returns:
+            List[List[int]]: A list of extraction groups, each of which is a list of integer
+                object indices that should be extracted together by core.skysub.local_skysub_extract
+        """
+        nobj = len(self.specobjs)
+
+        if model_full_slit:
+            return [list(range(nobj))]
+
+        # initialize adjacency matrix
+        adj = np.full((nobj, nobj), dtype=bool, fill_value=False)
+        ## build adjacency matrix
+        # adj[i, j] is True iff objects i and j are touching each other
+        for i in range(nobj):
+            left_edge_i = self.specobjs[i].TRACE_SPAT - self.specobjs[i].maskwidth - 1
+            righ_edge_i = self.specobjs[i].TRACE_SPAT + self.specobjs[i].maskwidth + 1
+            for j in range(i + 1, nobj):
+                left_edge_j = self.specobjs[j].TRACE_SPAT - self.specobjs[j].maskwidth - 1
+                righ_edge_j = self.specobjs[j].TRACE_SPAT + self.specobjs[j].maskwidth + 1
+
+                touch = (left_edge_j < righ_edge_i) & (left_edge_i < righ_edge_j)
+
+                if touch.any():
+                    adj[i, j] = True
+                    adj[j, i] = True
+
+        ## Find all connected components in the graph of objects.
+        # One call to DFS will visit every object it can that is "connected" to
+        # the starting object by the touching relation.
+        visited = [False]*nobj
+        groups = []
+        while not all(visited):
+            # pick a starting unvisited vertex
+            v = visited.index(False)
+            group = []
+            # DFS starting at v. Afterwards, group contains every object that
+            # is "connected" to v by the touching relation.
+            utils.DFS(v, visited, group, adj)
+            groups.append(group)
+
+        return groups
 
 
 def lst_to_array(lst, mask=None):
